#include "amxxmodule.h"
#include "SemiclipEx.h"
#include "pm_defs.h"
#include "pm_shared.h"

//AMXX forward

int g_fwd_PM_ShouldCollide = 0;

//here we backup the physent array
physent_t g_sBackupPhysent[MAX_PHYSENTS];
int g_iNumBackupPhysent;

int PM_ShouldCollide(int playerindex, int entindex)
{
	if(g_fwd_PM_ShouldCollide >= 1)
	{
		return MF_ExecuteForward(g_fwd_PM_ShouldCollide, (cell)playerindex, (cell)entindex);
	}
	return 1;
}

//Rebuild pmove's physents array now

void PM_RebuildPhysents(struct playermove_s *ppmove)
{
	int i;
	int iPlayerIndex = ppmove->player_index + 1;

	//backup physent
	g_iNumBackupPhysent = ppmove->numphysent;
	memcpy(g_sBackupPhysent, ppmove->physents, sizeof(physent_t) * ppmove->numphysent);

	//rebuild
	ppmove->numphysent = 0;

	for(i = 0; i < g_iNumBackupPhysent; i++)
	{
		if(!PM_ShouldCollide(iPlayerIndex, g_sBackupPhysent[i].info))
			continue;

		memcpy(&ppmove->physents[ppmove->numphysent ++], &g_sBackupPhysent[i], sizeof(physent_t));
	}
}

//Copy backup physents back to pmove

void PM_ResetPhysents(struct playermove_s *ppmove)
{
	ppmove->numphysent = g_iNumBackupPhysent;
	memcpy(ppmove->physents, g_sBackupPhysent, sizeof(physent_t)*g_iNumBackupPhysent);
}

//rebuild and recover pmove->physent here

void PM_Move(struct playermove_s *ppmove, int server)
{
	PM_RebuildPhysents(ppmove);

	RETURN_META(MRES_HANDLED);
}

void PM_Move_Post(struct playermove_s *ppmove, int server)
{
	PM_ResetPhysents(ppmove);

	RETURN_META(MRES_HANDLED);
}

//Register forward

static cell AMX_NATIVE_CALL SCE_RegisterForward( AMX * pAmx, cell * pParameters ) {
	int func = pParameters[1];
	int fwd;
	char *fwdname;

	fwdname = (char *)MF_GetAmxString( pAmx, pParameters[2], 0, NULL );

	if(func == func_PM_ShouldCollide)
	{
		fwd = MF_RegisterSPForwardByName(pAmx, fwdname, FP_CELL, FP_CELL, FP_DONE);
		if (fwd < 1)
		{
			MF_LogError(pAmx, AMX_ERR_NATIVE, "[SemiclipEx] Function not found: %s", fwdname);
			return 0;
		}
		g_fwd_PM_ShouldCollide = fwd;
	}
	else
	{
		MF_LogError(pAmx, AMX_ERR_NATIVE, "[SemiclipEx] Invalid Function index: %d", func);
	}

	return 0;
}

AMX_NATIVE_INFO natives[] = {
	{"SCE_RegisterForward",			SCE_RegisterForward},
	{NULL,				NULL},
};

void Engine_InstallHook(void);

void ServerDeactivate_Post(void)
{
	g_fwd_PM_ShouldCollide = 0;
}

void OnAmxxAttach(void)
{
	g_fwd_PM_ShouldCollide = 0;

	Engine_InstallHook();
	MF_AddNatives( natives );
}